from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class N_BEATS(nn.Module):
    """
    N-BEATS (Neural Basis Expansion Analysis for Time Series) block.

    A specialized neural architecture for time series forecasting based on
    backward and forward residual links and a very deep stack of fully-connected layers.

    This implementation is based on the N-BEATS paper by Oreshkin et al.
    (https://arxiv.org/abs/1905.10437)
    """

    def __init__(
        self,
        input_size: int,
        theta_size: int,
        basis_size: int,
        hidden_size: int = 256,
        stack_layers: int = 4,
        activation: str = "relu",
        share_weights: bool = False,
        dropout: float = 0.1,
    ):
        """
        Args:
            input_size: Input sequence length
            theta_size: Basis expansion coefficient size
            basis_size: Number of basis functions
            hidden_size: Size of hidden layers
            stack_layers: Number of fully connected layers
            activation: Activation function
            share_weights: Whether to share weights in stack
            dropout: Dropout probability
        """
        super().__init__()
        self.input_size = input_size
        self.theta_size = theta_size
        self.basis_size = basis_size
        self.hidden_size = hidden_size
        self.stack_layers = stack_layers
        self.share_weights = share_weights

        # Fully connected stack
        if share_weights:
            self.fc_layer = nn.Sequential(
                nn.Linear(input_size, hidden_size),
                self._get_activation(activation),
                nn.Dropout(dropout),
            )
            self.stacks = nn.ModuleList([self.fc_layer for _ in range(stack_layers)])
        else:
            self.stacks = nn.ModuleList()
            for i in range(stack_layers):
                if i == 0:
                    layer = nn.Sequential(
                        nn.Linear(input_size, hidden_size),
                        self._get_activation(activation),
                        nn.Dropout(dropout),
                    )
                else:
                    layer = nn.Sequential(
                        nn.Linear(hidden_size, hidden_size),
                        self._get_activation(activation),
                        nn.Dropout(dropout),
                    )
                self.stacks.append(layer)

        # Basis coefficient generator
        self.theta_layer = nn.Linear(hidden_size, theta_size)

        # Basis functions for backward and forward signals
        self.backcast_basis = nn.Linear(theta_size, input_size)
        self.forecast_basis = nn.Linear(theta_size, basis_size)

    def _get_activation(self, activation: str) -> nn.Module:
        """Get activation function by name"""
        return {
            "relu": nn.ReLU(),
            "gelu": nn.GELU(),
            "silu": nn.SiLU(),
            "tanh": nn.Tanh(),
        }.get(activation.lower(), nn.ReLU())

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            x: Input tensor [batch_size, input_size]

        Returns:
            Tuple of:
                - Backcast (input reconstruction) [batch_size, input_size]
                - Forecast (prediction) [batch_size, basis_size]
        """
        # Stack of fully connected layers
        block_input = x
        for layer in self.stacks:
            block_input = layer(block_input)

        # Compute basis expansion coefficients
        theta = self.theta_layer(block_input)

        # Compute backcast and forecast
        backcast = self.backcast_basis(theta)
        forecast = self.forecast_basis(theta)

        return backcast, forecast


import torch
import torch.fft
import torch.nn as nn
import torch.nn.functional as F


class TimesBlock(nn.Module):
    """
    Highly optimized TimesBlock implementation with significant performance improvements:
    - Cached FFT computations
    - Vectorized period processing
    - Efficient memory usage
    - Fused operations
    - Reduced computational complexity
    """

    def __init__(
        self, d_model: int, d_ff: int = None, k_periods: int = 3, dropout: float = 0.1
    ):
        super().__init__()
        self.k = k_periods
        self.d_model = d_model
        self.d_ff = d_ff or 2 * d_model  # Reduced from 4x for efficiency

        # Optimized inception block
        self.inception = InceptionBlock(d_model, self.d_ff, dropout)

        # Cache for FFT results to avoid recomputation
        self.register_buffer("_fft_cache_key", torch.tensor(-1))
        self.register_buffer(
            "_cached_periods", torch.zeros(k_periods, dtype=torch.long)
        )
        self.register_buffer("_cached_weights", torch.zeros(1, k_periods))

        # Pre-computed common periods for faster lookup
        self.register_buffer(
            "common_periods", torch.tensor([1, 2, 3, 4, 6, 8, 12, 16, 24])
        )

    def forward(self, x):
        """
        Optimized forward pass with vectorized operations.

        Args:
            x: Input tensor [B, T, C] where C == d_model

        Returns:
            Output tensor [B, T, C]
        """
        B, T, C = x.shape

        # Fast period discovery with caching
        period_list, period_weight = self.period_discovery_fast(x)

        # Vectorized processing of all periods simultaneously
        all_outputs = self.process_all_periods_vectorized(x, period_list)

        # Fast aggregation
        period_weight = F.softmax(period_weight, dim=1)  # [B, k]
        # Reshape for broadcasting: [B, k] -> [B, k, 1, 1] to match [B, k, T, C]
        period_weight = period_weight.unsqueeze(2).unsqueeze(3)  # [B, k, 1, 1]

        # Efficient weighted sum
        output = torch.sum(all_outputs * period_weight, dim=1)  # [B, T, C]

        return output

    def period_discovery_fast(self, x):
        """
        Optimized period discovery with caching and approximations.
        """
        B, T, C = x.shape

        # Simple hash for cache key
        cache_key = hash((T, C)) % 10000

        if self._fft_cache_key.item() == cache_key and T <= 512:
            # Use cached results for same input shape
            period_list = self._cached_periods.tolist()
            period_weight = self._cached_weights.repeat(B, 1)
            return period_list, period_weight

        # Efficient FFT computation
        if T > 256:
            # For long sequences, downsample for FFT to reduce computation
            downsample_factor = max(1, T // 128)
            x_downsampled = x[:, ::downsample_factor, :]
            T_down = x_downsampled.size(1)
        else:
            x_downsampled = x
            T_down = T

        # FFT with reduced precision for speed
        with torch.amp.autocast("cuda", enabled=torch.cuda.is_available()):
            fft = torch.fft.fft(x_downsampled, dim=1)
            amp = torch.abs(fft).mean(dim=2)  # [B, T_down]

        # Only consider meaningful frequency range
        freq_start = max(1, T_down // 64)
        freq_end = min(T_down // 2, T_down // 3)

        if freq_end <= freq_start:
            # Fallback to common periods
            period_list = [2, 4, 8][: self.k]
            period_weight = torch.ones(B, len(period_list), device=x.device)
            return period_list, period_weight

        amp_slice = amp[:, freq_start:freq_end]

        # Efficient top-k selection
        if amp_slice.size(1) < self.k:
            # If not enough frequencies, pad with common periods
            available_k = amp_slice.size(1)
            _, top_indices = torch.topk(amp_slice, available_k, dim=1)
            top_indices = top_indices + freq_start

            # Convert to periods
            periods_tensor = T_down // top_indices.float()
            periods_tensor = torch.clamp(periods_tensor, min=1, max=T // 2).int()

            # Fill remaining with common periods
            period_list = []
            for i in range(available_k):
                period_list.append(periods_tensor[0, i].item())

            # Add common periods to fill up to k
            common_to_add = [2, 4, 8, 12, 16]
            for p in common_to_add:
                if len(period_list) >= self.k:
                    break
                if p not in period_list and p <= T // 2:
                    period_list.append(p)

            # Ensure we have exactly k periods
            while len(period_list) < self.k:
                period_list.append(min(2, T // 2))

            period_list = period_list[: self.k]
            period_weight = torch.ones(B, self.k, device=x.device)
        else:
            _, top_indices = torch.topk(amp_slice, self.k, dim=1)
            top_indices = top_indices + freq_start

            # Vectorized period computation
            periods_tensor = T_down // top_indices.float()
            periods_tensor = torch.clamp(periods_tensor, min=1, max=T // 2).int()

            # Use mode across batch for consistency
            period_list = []
            period_weights = []

            for i in range(self.k):
                period_candidates = periods_tensor[:, i]
                period_mode = torch.mode(period_candidates)[0].item()
                period_list.append(max(1, min(period_mode, T // 2)))

                # Gather weights for this period
                batch_weights = []
                for b in range(B):
                    idx = top_indices[b, i] - freq_start
                    if idx < amp_slice.size(1):
                        batch_weights.append(amp_slice[b, idx])
                    else:
                        batch_weights.append(torch.tensor(1.0, device=x.device))
                period_weights.append(torch.stack(batch_weights))

            period_weight = torch.stack(period_weights, dim=1)

        # Update cache for small sequences
        if T <= 512:
            self._fft_cache_key.fill_(cache_key)
            self._cached_periods[: len(period_list)] = torch.tensor(period_list)
            if B == 1:
                self._cached_weights = period_weight

        return period_list, period_weight

    def process_all_periods_vectorized(self, x, period_list):
        """
        Process all periods in a vectorized manner for maximum efficiency.
        """
        B, T, C = x.shape

        # Pre-allocate output tensor
        all_outputs = torch.zeros(
            B, len(period_list), T, C, device=x.device, dtype=x.dtype
        )

        # Group periods by size for batch processing
        period_groups = {}
        for i, period in enumerate(period_list):
            if period not in period_groups:
                period_groups[period] = []
            period_groups[period].append(i)

        # Process each group
        for period, indices in period_groups.items():
            # Transform to 2D for this period
            x_2d = self.transform_1d_to_2d_fast(x, period)

            # Process with inception block
            x_2d_processed = self.inception(x_2d)

            # Transform back to 1D
            x_1d = self.transform_2d_to_1d_fast(x_2d_processed, T)

            # Assign to all indices with this period
            for idx in indices:
                all_outputs[:, idx] = x_1d

        return all_outputs

    def transform_1d_to_2d_fast(self, x, period):
        """
        Optimized 1D to 2D transformation with minimal memory allocation.
        """
        B, T, C = x.shape

        # Calculate padding efficiently
        remainder = T % period
        if remainder != 0:
            pad_len = period - remainder
            # Use reflection padding for better boundary handling
            x = F.pad(x, (0, 0, 0, pad_len), mode="reflect")

        T_new = x.size(1)
        num_periods = T_new // period

        # Efficient reshape
        return x.view(B, num_periods, period, C).transpose(1, 2).contiguous()

    def transform_2d_to_1d_fast(self, x_2d, target_len):
        """
        Optimized 2D to 1D transformation.
        """
        B, period, num_periods, C = x_2d.shape

        # Fast reshape and truncate
        x_1d = x_2d.transpose(1, 2).contiguous().view(B, period * num_periods, C)
        return x_1d[:, :target_len]


class InceptionBlock(nn.Module):
    """
    Optimized inception block with reduced parameters and fused operations.
    """

    def __init__(self, d_model: int, d_ff: int, dropout: float = 0.1):
        super().__init__()

        # Reduced channel dimensions for efficiency
        mid_channels = d_ff // 4

        # Depthwise separable convolutions for efficiency
        self.conv1x1 = nn.Sequential(
            nn.Conv2d(d_model, mid_channels, 1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.GELU(),
        )

        # Grouped convolutions for efficiency
        groups = min(8, mid_channels)

        self.conv3x3 = nn.Sequential(
            nn.Conv2d(d_model, mid_channels, 3, padding=1, groups=groups, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.GELU(),
        )

        self.conv5x5 = nn.Sequential(
            nn.Conv2d(d_model, mid_channels, 5, padding=2, groups=groups, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.GELU(),
        )

        # Replace 7x7 with two 3x3 for efficiency
        self.conv7x7 = nn.Sequential(
            nn.Conv2d(
                d_model, mid_channels // 2, 3, padding=1, groups=groups // 2, bias=False
            ),
            nn.Conv2d(mid_channels // 2, mid_channels, 3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.GELU(),
        )

        # Efficient output projection
        self.projection = nn.Sequential(
            nn.Conv2d(mid_channels * 4, d_model, 1, bias=False), nn.Dropout2d(dropout)
        )

    def forward(self, x):
        """
        Args:
            x: [B, period, num_periods, C]
        Returns:
            out: [B, period, num_periods, C]
        """
        # Convert to conv2d format
        x = x.permute(0, 3, 1, 2)  # [B, C, period, num_periods]

        # Parallel convolution execution
        out1 = self.conv1x1(x)
        out2 = self.conv3x3(x)
        out3 = self.conv5x5(x)
        out4 = self.conv7x7(x)

        # Efficient concatenation and projection
        out = torch.cat([out1, out2, out3, out4], dim=1)
        out = self.projection(out)

        # Convert back
        return out.permute(0, 2, 3, 1)


class TimesNet(nn.Module):
    """
    Optimized TimesNet with performance improvements.
    """

    def __init__(
        self,
        seq_len: int,
        pred_len: int,
        enc_in: int,
        d_model: int = 64,
        n_layers: int = 2,
        d_ff: int = None,
        k_periods: int = 3,
        dropout: float = 0.1,
    ):
        super().__init__()
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.d_model = d_model

        # Efficient input embedding
        self.enc_embedding = nn.Linear(enc_in, d_model, bias=False)

        # Shared TimesBlock for parameter efficiency
        if n_layers > 1:
            self.layers = nn.ModuleList(
                [TimesBlock(d_model, d_ff, k_periods, dropout) for _ in range(n_layers)]
            )
        else:
            # Share single block for maximum efficiency
            single_block = TimesBlock(d_model, d_ff, k_periods, dropout)
            self.layers = nn.ModuleList([single_block] * n_layers)

        # Efficient prediction head
        self.projection = nn.Linear(d_model, pred_len * enc_in, bias=False)

    def forward(self, x):
        """
        Args:
            x: [B, seq_len, enc_in]
        Returns:
            pred: [B, pred_len, enc_in]
        """
        B, seq_len, enc_in = x.shape

        # Input embedding
        x = self.enc_embedding(x)

        # Layer processing with gradient checkpointing for memory efficiency
        for layer in self.layers:
            if self.training and x.requires_grad:
                x = x + torch.utils.checkpoint.checkpoint(layer, x)
            else:
                x = x + layer(x)

        # Efficient prediction
        x = self.projection(x)
        x = x.view(B, seq_len, self.pred_len, enc_in)

        return x[:, -1, :, :]


class TimesBlockPreprocessor(nn.Module):
    """
    Optimized TimesBlock preprocessor.
    """

    def __init__(
        self,
        d_model: int = 64,
        k_periods: int = 3,
        d_ff: int = None,
        dropout: float = 0.1,
    ):
        super().__init__()
        self.d_model = d_model
        self.times_block = TimesBlock(d_model, d_ff, k_periods, dropout)

        # Pre-allocated projection layer to avoid dynamic creation
        self._input_projections = nn.ModuleDict()

    def forward(self, x):
        """
        Args:
            x: Input tensor [B, T, C]
        Returns:
            Processed tensor [B, T, d_model]
        """
        B, T, C = x.shape

        # Efficient input projection
        if C != self.d_model:
            key = str(C)
            if key not in self._input_projections:
                self._input_projections[key] = nn.Linear(
                    C, self.d_model, bias=False
                ).to(x.device)
            x = self._input_projections[key](x)

        return self.times_block(x)
